#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === AsciiParadise::ParseStaticAscii
#
# This class can be used to parse a static ascii file, that is,
# specifically, a file that can be used for saying some text
# embedded into some ASCII art.
#
# The old name was Speak, but the name was IMO confusing, so the
# class was renamed during August 2019.
# =========================================================================== #
# require 'ascii_paradise/ascii_say/parse_static_ascii.rb'
# =========================================================================== #
require 'ascii_paradise/constants/constants.rb'

module AsciiParadise

class ParseStaticAscii # === AsciiParadise::ParseStaticAscii

  # ========================================================================= #
  # Note that the various .ascii files will keep the entries for
  # message_sentinel, message_callout_symbol and
  # message_callout_sentinel within that respective file.
  #
  # This explains why eval is used - we will eval on the
  # header.
  # ========================================================================= #
  attr_accessor :message_sentinel
  attr_accessor :message_callout_symbol
  attr_accessor :message_callout_sentinel

# ========================================================================= #
  # === initialize
  #
  # The first argument must be the path to the template file in question.
  # ========================================================================= #
  def initialize(
      this_template_file,
      run_already = true
    )
    reset
    set_path_to_the_template_file(
      this_template_file
    )
    run if run_already
  end

  # ========================================================================= #
  # === set_path_to_the_template_file
  #
  # Set to the path where the template file at hand is kept. This is
  # important, as otherwise the class will not be able to work properly.
  # ========================================================================= #
  def set_path_to_the_template_file(i)
    @path_to_the_template_file = i
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    @alignment = :center
  end

  # ========================================================================= #
  # === set_body
  # ========================================================================= #
  def set_body(i)
    @body = i
  end

  # ========================================================================= #
  # === parse
  #
  # Input to this method should be an Array.
  # ========================================================================= #
  def parse(i)
    array = i.split(/^---\n/) # Split on lines starting with --- and then a newline.
    # ======================================================================= #
    # === Get rid of comments. This seems to be unnecessary, though.
    #     It is retained here in the event that we may wish to support
    #     comments in these files one day.
    # ======================================================================= #
    # array.reject! {|entry| entry.strip.start_with? '#' }
    # ======================================================================= #
    # Our .ascii file typically has a header and a body section. Obtain
    # each section next.
    # ======================================================================= #
    header, body = array
    binding.eval(header) # <- The header includes 3 instance variables.
    set_body(body) # Set the body.
  end

  # ========================================================================= #
  # === print
  # ========================================================================= #
  def print(i = nil)
    if i
      # ===================================================================== #
      # Create a clean slate for the body buffer. We will overwrite
      # it at a later time.
      #
      # Note that message_sentinel is typically the character 'x',
      # so this character will be removed in the following line.
      # ===================================================================== #
      body_buffer = @body.tr(message_sentinel, ' ')
      # ===================================================================== #
      # Handle the case when there is a message_callout_sentinel deined.
      # ===================================================================== #
      if message_callout_sentinel
        # =================================================================== #
        # In this case, replace one with the other.
        # =================================================================== #
        body_buffer = body_buffer.tr(
          message_callout_sentinel, message_callout_symbol
        )
      end
      body_buffer = body_buffer.each_line.to_a

      line_idx_stack = @body.each_line.each_with_index.select { |line, idx|
        line[/#{Regexp.escape(message_sentinel)}/]
      }
      array_message_stack = i.split(' ')
      # ===================================================================== #
      # Obtain the current message next:
      # ===================================================================== #
      current_message = array_message_stack.pop
      until (current_message.nil? && array_message_stack.empty?) || line_idx_stack.empty?
        current_line, idx = line_idx_stack.pop
        message_buffer = ''
        # =================================================================== #
        # TODO making an assumption that lines only have one place
        # to replace text and it is contiguous.
        # =================================================================== #
        current_line_capacity = current_line[/#{Regexp.escape(message_sentinel)}+/].length
        # =================================================================== #
        # Until no more messages or buffer would overflow
        # =================================================================== #
        until current_message.nil? || (current_message.length + (message_buffer.length == 0 ? 0 : message_buffer.length + 1)) > current_line_capacity
          # ================================================================= #
          # Add the message to the buffer.
          # ================================================================= #
          if message_buffer.length > 0
            message_buffer = [current_message, message_buffer].join(' ')
          else
            message_buffer = current_message
          end
          current_message = array_message_stack.pop
        end
        # =================================================================== #
        # Pad the buffer to capacity.
        # =================================================================== #
        case alignment # case tag.
        # =================================================================== #
        # === :left
        #
        # This entry point will shift the message-buffer by a bit.
        # =================================================================== #
        when :left
          message_buffer << ' ' * (current_line_capacity - message_buffer.length)
        # =================================================================== #
        # === :right
        #
        # Align to the right hand side of the ASCII picture.
        # =================================================================== #
        when :right
          message_buffer.insert 0, ' ' * (current_line_capacity - message_buffer.length)
        else
          # ================================================================= #
          # Determine the half.
          # ================================================================= #
          half = (current_line_capacity - message_buffer.length) / 2.0
          message_buffer = message_buffer.dup if message_buffer.frozen?
          message_buffer.insert(0, ' ' * half.floor)
          message_buffer << ' ' * half.ceil
        end
        # =================================================================== #
        # Flush the buffer.
        # =================================================================== #
        line = current_line.gsub(
          /#{Regexp.escape(message_sentinel * current_line_capacity)}/,
          message_buffer
        )
        body_buffer[idx] = line
      end
      if array_message_stack.empty?
        puts body_buffer.join
      else # The error may be raised when we have too many characters.
        n_characters = array_message_stack.join(' ').size
        raise "Too many characters given. Should not give "\
              "more than #{n_characters} characters."
      end
    else
      body = @body.tr(message_sentinel, ' ')
      if message_callout_sentinel
        body.tr!(message_callout_sentinel, ' ')
      end
      e body
    end
  end; alias ee  print # === ee
       alias say print # === say

  # ========================================================================= #
  # === alignment?
  # ========================================================================= #
  def alignment?
    @alignment
  end; alias alignment alignment? # === alignment

  # ========================================================================= #
  # === alignment=
  # ========================================================================= #
  def alignment=(i)
    @alignment = i
  end

  # ========================================================================= #
  # === parse_the_template_file
  # ========================================================================= #
  def parse_the_template_file(i = @path_to_the_template_file)
    parse(
      File.read(i)
    )
  end

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    parse_the_template_file
  end

end

  # ========================================================================= #
  # === AsciiParadise.register
  #
  # The second argument to this method is the path to the .ascii file
  # in question, the template that we will use for the speak-bubble.
  # ========================================================================= #
  def self.register(name, template)
    AsciiParadise.const_set(
      name, AsciiParadise::ParseStaticAscii.new(template)
    )
  end

  # ========================================================================= #
  # Register the four classes that are available as of yet.
  #
  # You can invoke them by issuing this here:
  #
  #   AsciiParadise::CatInTheHat.say 'I want a mouse!'
  #   AsciiParadise::Buddha.say 'Love <3'
  #   AsciiParadise::Wizard.say 'You shall not pass!'
  #   AsciiParadise::Alien.say 'hello world'
  #   AsciiParadise::Knight.say 'Charge!!!'
  #   AsciiParadise::SexyAngel.say 'Love <3'
  #   AsciiParadise::Cowsay.say 'I am a cow - moo moo moooooOOO!'
  #
  # ========================================================================= #
  ARRAY_SAY_CONSTANTS.each {|name|
    camel_cased = name
    if camel_cased.include? '_'
      camel_cased = camel_cased.split('_').map {|entry| entry.capitalize }.join
    else
      camel_cased = camel_cased.capitalize
    end
    # ======================================================================= #
    # Register the path here. Note that we hardcode the file extension
    # to .ascii text files, which are essentially .ascii files.
    # ======================================================================= #
    AsciiParadise.register(
      camel_cased.to_sym,
      File.join(template_directory?, "#{name}.ascii")
    )
  }

end