#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# AsciiParadise::AsciiTable::Cell
# =========================================================================== #
module AsciiParadise

class AsciiTable

class Cell

  # attr_reader :width  # Cell width.
  attr_reader :colspan  # Column span.

  # ========================================================================= #
  # === initialize
  #
  # Initialize with _options_.
  # ========================================================================= #
  def initialize(options = nil)
    reset
    @value, options = options, {} unless Hash === options
    @value = options.fetch :value, value
    @alignment = options.fetch :alignment, nil
    @alignment = :center if @alignment == :middle
    @colspan = options.fetch :colspan, 1
    @width = options.fetch :width, @value.to_s.size
    @index = options.fetch :index
    @table = options.fetch :table
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    @value = 0
  end

  # ========================================================================= #
  # === value?
  # ========================================================================= #
  def value?
    @value
  end; alias value value? # === value

  # ========================================================================= #
  # === alignment=
  # ========================================================================= #
  def alignment=(i)
    supported_alignments = %w( left center right )
    case i
    when :middle
      i = :center
    end
    if supported_alignments.include?(i.to_s)
      @alignment = i
    else
      raise "Aligment must be one of: #{supported_alignments.join(' ')}"
    end
  end

  # ========================================================================= #
  # === lines
  # ========================================================================= #
  def lines
    @value.to_s.split(/\n/)
  end

  # ========================================================================= #
  # === value_for_column_width_recalc
  #
  # Returns the longest line in the cell and removes all ANSI escape
  # sequences (e.g. color).
  # ========================================================================= #
  def value_for_column_width_recalc
    lines.map{ |s| escape(s) }.max_by{ |s| s.size }
  end

  # ========================================================================= #
  # === width
  #
  # Returns the width of this cell
  # ========================================================================= #
  def width
    padding = (colspan - 1) * @table.cell_spacing
    inner_width = (1..@colspan).to_a.inject(0) { |w, counter|
      w + @table.column_width(@index + counter - 1)
    }
    return inner_width + padding
  end

  # ========================================================================= #
  # === escape
  #
  # This method removes all ANSI escape sequences (e.g. color).
  # ========================================================================= #      
  def escape(i)
    i.to_s.gsub(/\x1b(\[|\(|\))[;?0-9]*[0-9A-Za-z]/, '').
           gsub(/\x1b(\[|\(|\))[;?0-9]*[0-9A-Za-z]/, '').
           gsub(/[\x03|\x1a]/, '')
  end

  # ========================================================================= #
  # === alignment?
  # ========================================================================= #
  def alignment?
    ! @alignment.nil?
  end

  # ========================================================================= #
  # === alignment
  # ========================================================================= #
  def alignment
    @alignment || :left
  end

  # ========================================================================= #
  # === render
  #
  # Render the cell.
  #
  # alignment can be :left or :right or :center.
  # ========================================================================= #
  def render(line = 0)
    left  = ' ' * @table.style.padding_left
    right = ' ' * @table.style.padding_right
    render_width = lines[line].to_s.size - escape(lines[line]).size + width
    line = "#{left}#{lines[line]}#{right}"
    horizontal_length = render_width + @table.cell_padding
    case alignment
    when :left
      line.ljust(horizontal_length)
    when :right
      line.rjust(horizontal_length)
    when :center
      line.center(horizontal_length)
    end
  end; alias to_s render # === to_s

end; end; end