#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === AsciiParadise::Table::Style
#
# A Style object holds all the formatting information for a Table object.
# 
# To create a table with a certain style, use either the constructor 
# option :style, the Table#style object or the Table#style= method.
#
# All these examples have the same effect:
#
#   # by constructor
#   @table = Table.new(:style => {:padding_left => 2, :width => 40})
#
#   # by object
#   @table.style.padding_left = 2
#   @table.style.width = 40
#
#   # by method
#   @table.style = {:padding_left => 2, :width => 40}
# 
# To set a default style for all tables created afterwards,
# use Style.defaults=
#
#     AsciiParadise::AsciiTable::Style.defaults = {:width => 80}
#
# =========================================================================== #
module AsciiParadise

class AsciiTable

class Style

  attr_accessor :border_x
  attr_accessor :border_y
  attr_accessor :border_i
  attr_accessor :padding_left
  attr_accessor :padding_right

  # ========================================================================= #
  # === DEFAULTS
  #
  # Three different tokens are possible as border style, which are
  # '-','|' and '+'.
  # ========================================================================= #
  DEFAULTS = {
    border_x:      '-',
    border_y:      '|',
    border_i:      '+', # 3 borders.
    padding_left:  1,
    padding_right: 1,
    width:         nil
  }

  # ========================================================================= #
  # === initialize
  # ========================================================================= #    
  def initialize(i = {})
    apply(
      self.class.defaults.merge(i)
    )
  end

  # ========================================================================= #
  # === apply
  #
  # The input to this method should be an Array.
  # ========================================================================= #
  def apply(i)
    i.each { |m, v| __send__ "#{m}=", v }
  end

  # ========================================================================= #
  # === Style.defaults
  # ========================================================================= #
  def self.defaults # Can be called like so: Style.defaults
    DEFAULTS
  end

  # ========================================================================= #
  # === Style.defaults=
  # ========================================================================= #
  def self.defaults=(i)
    defaults.merge(i) # Call the above method here.
  end

  # ========================================================================= #
  # === width?
  # ========================================================================= #
  def width?
    @width
  end; alias width width? # === width

  # ========================================================================= #
  # === width=
  # ========================================================================= #
  def width=(i)
    @width = i
  end

end; end; end