#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
module AsciiParadise

class AsciiTable

  # ========================================================================= #
  # === initialize
  #
  # Generates a ASCII table with the given options.
  # ========================================================================= #
  def initialize(
      i = {}, &block
    )
    reset
    self.style    = i.fetch :style, {}
    self.headings = i.fetch :headings, []
    self.rows     = i.fetch :rows, []
    self.title    = i.fetch :title, nil
    yield_or_eval(&block) if block
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    # === @column_widths
    @column_widths = []
    # === @result
    @result = nil
  end

  # ========================================================================= #
  # === headings?
  #
  # Reader method for @headings.
  # ========================================================================= #
  def headings?
    @headings
  end; alias headings headings? # === headings

  # ========================================================================= #
  # === style=                                                  (style tag)
  #
  # Set the style setter. We do this by calling apply() on the style object.
  #
  # Specific example how to use this:
  #   t.style = {:padding_left => 2, :width => 80}
  # ========================================================================= #
  def style=(i)
    style.apply(i) # The method style() returns a @style object - see below.
  end

  # ========================================================================= #
  # === style
  # ========================================================================= #
  def style
    @style ||= Style.new # Can be found in the file style.rb.
  end

  # ========================================================================= #
  # === align_column
  #
  # Align column _n_ to the given _alignment_ of :center, :left, or
  # :right.
  # ========================================================================= #
  def align_column(n, alignment)
    case alignment
    when :middle
      alignment = :center
    end
    r = rows
    column(n).each_with_index { |col, i|
      cell = r[i][n]
      next unless cell
      cell.alignment = alignment unless cell.alignment?
    }
  end

  # ========================================================================= #
  # === add_row                                       (row tag, append tag)
  #
  # Add a row. We add either a separator, or a new row.
  #
  # You can also use the << operator for this method.
  # ========================================================================= #
  def add_row(array)
    if array == :separator
      row = Separator.new(self)
    else
      row = Row.new(self, array)
    end 
    @rows << row
    recalc_column_widths(row)
  end; alias <<     add_row # === <<
       alias append add_row # === append

  # ========================================================================= #
  # === add
  # ========================================================================= #
  def add(i = :separator)
    case i
    when :separator
      add_separator
    else # Else assume a row.
      add_row(i)
    end
  end

  # ========================================================================= #
  # === add_separator
  #
  # Add a separator. A separator has ---- and +.
  # ========================================================================= #
  def add_separator
    add_row(:separator)
  end

  # ========================================================================= #
  # === cell_spacing
  # ========================================================================= #
  def cell_spacing
    cell_padding + style.border_y.length
  end

  # ========================================================================= #
  # === cell_padding
  #
  # Return the total padding used here.
  # ========================================================================= #
  def cell_padding
    style.padding_left + style.padding_right
  end

  # ========================================================================= #
  # === column
  #
  # Return column _n_.
  # ========================================================================= #
  def column(n, method = :value, array = rows)
    array.map { |row| 
     cell = row[n]
     cell && method ? cell.__send__(method) : cell
    }.compact 
  end

  # ========================================================================= #
  # === column_with_headings
  #
  # Return _n_ column including headings.
  # ========================================================================= #
  def column_with_headings(n, method = :value)
    column(n, method, headings_with_rows)
  end

  # ========================================================================= #
  # === columns
  #
  # Return columns.
  # ========================================================================= #
  def columns
    (0...number_of_columns).map { |n| column n } 
  end

  # ========================================================================= #
  # === column_width
  #
  # Return length of column _n_.
  # ========================================================================= #
  def column_width(n)
    width = @column_widths[n] || 0
    width + additional_column_widths[n].to_i
  end; alias length_of_column column_width # for legacy support

  # ========================================================================= #
  # === number_of_columns
  #
  # Return total number of columns available.
  # ========================================================================= #
  def number_of_columns
    headings_with_rows.map { |r| r.cells.size }.max
  end

  # ========================================================================= #
  # === headings=
  #
  # Set the headings
  # ========================================================================= #
  def headings=(i)
    @headings = Row.new(self, i)
    recalc_column_widths @headings
  end

  # ========================================================================= #
  # === render
  #
  # Render the table.
  # ========================================================================= #
  def render
    separator = Separator.new(self)
    buffer = [separator]
    unless @title.nil?
      buffer << Row.new(self, [title_cell_options])
      buffer << separator
    end
    unless @headings.cells.empty?
      buffer << @headings
      buffer << separator
    end
    buffer += @rows
    buffer << separator
    buffer.map { |r| r.render }.join("\n")
  end; alias to_s render # to_s tag.

  # ========================================================================= #
  # === rows
  #
  # Return rows without separator rows.
  # ========================================================================= #
  def rows
    @rows.reject { |row| row.is_a? Separator }
  end

  # ========================================================================= #
  # === rows
  #
  # Input to this should be an array.
  # ========================================================================= #
  def rows=(i)
    @rows = []
    i.each { |entry| self << entry }
  end

  # ========================================================================= #
  # === title=                                                  (title tag)
  #
  # Set the title here.
  # ========================================================================= #
  def title=(i)
    @title = i
    recalc_column_widths( Row.new(self, [title_cell_options]) )
  end; alias header=   title= # === header=
       alias set_title title= # === set_title

  # ========================================================================= #
  # === ==
  #
  # Check if _other_ is equal to self. _other_ is considered equal
  # if it contains the same headings and rows.
  # ========================================================================= #
  def ==(other)
    if other.respond_to? :render and other.respond_to? :rows
      self.headings == other.headings and self.rows == other.rows
    end
  end

  # ========================================================================= #
  # === yield_or_eval                                             (run tag)
  # ========================================================================= #
  def yield_or_eval(&block)
    return unless block
    if block.arity > 0
      yield self
    else
      @result = self.instance_eval(&block)
    end
    return @result
  end; alias display yield_or_eval # === display
       alias run     yield_or_eval # === run

  private

  # ========================================================================= #
  # === columns_width
  # ========================================================================= #
  def columns_width
    @column_widths.inject(0) { |s, i|
      s + i + cell_spacing
    } + style.border_y.length
  end

  # ========================================================================= #
  # === additional_column_widths
  # ========================================================================= #
  def additional_column_widths
    return [] if style.width.nil?
    spacing = style.width - columns_width
    if spacing < 0
      raise "Table width exceeds wanted width of `#{style.width}` characters."
    else
      per_col = spacing / number_of_columns
      array = (1...number_of_columns).to_a.map { |i| per_col }
      other_cols = array.inject(0) { |s, i| s + i }
      array << spacing - other_cols
      array
    end
  end

  # ========================================================================= #
  # === recalc_column_widths
  # ========================================================================= #
  def recalc_column_widths(row)
    return if row.is_a? Separator
    i = 0
    row.cells.each { |cell|
      colspan = cell.colspan
      cell_value = cell.value_for_column_width_recalc
      colspan.downto(1) { |j|
        cell_length = cell_value.to_s.length
        if colspan > 1
          spacing_length = cell_spacing * (colspan - 1)
          length_in_columns = (cell_length - spacing_length)
          cell_length = (length_in_columns.to_f / colspan).ceil
        end
        if @column_widths[i].to_i < cell_length
          @column_widths[i] = cell_length
        end
        i += 1
      }
    }
  end

  # ========================================================================= #
  # === result?
  # ========================================================================= #
  def result?
    @result
  end; alias result result? # === result

  # ========================================================================= #
  # === title?
  # ========================================================================= #
  def title?
    @title
  end; alias title title? # === title

  # ========================================================================= #
  # === headings_with_rows
  #
  # Return headings combined with rows.
  # ========================================================================= #
  def headings_with_rows
    [@headings] + rows
  end

  # ========================================================================= #
  # === title_cell_options
  # ========================================================================= #
  def title_cell_options
    {
      :value     => @title,
      :alignment => :center,
      :colspan   => number_of_columns
    }
  end

end; end