#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# This file makes available some sort of "interactive menu", that will
# allow a user to run any AsciiParadise component that is animated.
#
# After it has finished running, the user may then quit or continue
# with the next animation in line. Partial input will also work.
# =========================================================================== #
# require 'ascii_paradise/interactive/interactive.rb'
# =========================================================================== #
require 'ascii_paradise/colours/colours.rb'
require 'ascii_paradise/toplevel_methods/available_ascii_components.rb'
require 'ascii_paradise/toplevel_methods/register_sigint.rb'
require 'ascii_paradise/requires/require_animations_files.rb'

module AsciiParadise

  # ========================================================================= #
  # === AsciiParadise.interactive
  #
  # Use this method if you want to ask interactively what component
  # we should run next.
  # ========================================================================= #
  def self.interactive
    @show_this_message_on_exit = 'We will exit now, as requested.'
    register_sigint
    initialize_animated_ascii_components
    ask_the_user_which_component_should_be_run_next
  end

  # ========================================================================= #
  # === AsciiParadise.show_interactive_help                       (help tag)
  #
  # This method will allow the user to select the relevant component
  # to run.
  # ========================================================================= #
  def self.show_interactive_help
    e 'These interactive components are available:'
    e
    # ======================================================================= #
    # === Get a handle on all the ASCII components of this project next
    #
    # This requires that we initiate these components first when running
    # this interactive-component, which has been done before via
    # initialize_animated_ascii_components().
    # ======================================================================= #
    components = return_all_animated_ascii_components
    components.each_with_index {|entry, index|
      index += 1
      e index.to_s.rjust(4)+' '+Colours.simp(entry)
    }
    e
    e 'Please input the positional number only, such as "50", without '\
      'the "" quotes.'
    e
    e 'If you wish to use colours-gem, simply input the colour, '\
      'then press enter.'
    e 'All components from that point onwards will use this colour.'
    e 'Use "clear" to clear it again.'
    # e 'You can input either the full name, a partial name, or the '\
    #   'positional number.'
    # e
    # e 'So for instance, if you have this line here "50 AsciiParadise::Worm", '\
    #   'then'
    # e 'you can invoke this either by typing 50, by typing worm or Worm '\
    #   'but also'
    # e 'by typing AsciiParadise::Worm or AsciiParadise.Worm.'
  end

  # ========================================================================= #
  # === AsciiParadise.ask_the_user_which_component_should_be_run_next
  #
  # We will ask the user which component should be run next.
  # ========================================================================= #
  def self.ask_the_user_which_component_should_be_run_next
    e
    e 'Input your command next. Input "help" or "h" to get all options and'
    e '"exit" or "q" to exit from this user input loop again.'
    e 
    return_value = loop {
      user_input = $stdin.gets.chomp.strip
      # ===================================================================== #
      # === Downcase it next
      # ===================================================================== #
      user_input.downcase!
      case user_input # case tag
      # ===================================================================== #
      # === clear
      # ===================================================================== #
      when /^clear$/i
        e 'Clearing colour assignments next.'
        clear_colours
      # ===================================================================== #
      # === help
      # ===================================================================== #
      when /^-?-?help$/i,'h'
        show_interactive_help
      # ===================================================================== #
      # === exit
      # ===================================================================== #
      when 'exit','q'
        e 'Exiting now.'
        exit
      else # Else the user did input something. We will evaluate this.
        if ::Colours.does_include? user_input
          e 'Now using Colours. '+user_input+' colour.'
          @colour = user_input
        else
          components = ::AsciiParadise.return_all_animated_ascii_components
          result = nil
          # possible_match =
          #   result.map {|entry| entry.to_s }.select {|entry| entry.include? user_input }
          # =================================================================== #
          # === First check if it is a number
          # =================================================================== #
          if user_input =~ /^\d+$/
            result = components[user_input.to_i - 1]
          # =================================================================== #
          # === Check if the input is part of the component
          # =================================================================== #
          # elsif !possible_match.empty? # Not enable dfor now.
          # else # Else we did not find anything.
          end
          if result
            _ = nil
            if @colour
              _ = '--use-this-colour='+@colour.to_s
            end
            result.new(_) # Instantiate it here.
          end
        end
      end
    }
    return return_value
  end

end

if __FILE__ == $PROGRAM_NAME
  AsciiParadise.interactive
end # ainteractive