#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === AsciiParadise::Sparky
# =========================================================================== #
require 'ascii_paradise/sparky/constants.rb'

module AsciiParadise

class Sparky

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(i)
    reset
    unless i.empty?
      @original_numbers = i
      numbers  = normalize_numbers(i)
      one_step = step_height(numbers)
      @ticks = numbers.map { |n|
        index = (n / one_step).to_i
        PROGRESSION_BLOCKS[index]
      }
    end
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    @ticks = []
  end

  # ========================================================================= #
  # === format
  #
  # Formats all the ticks in the given Sparkline with a given block,
  # returns itself.
  #
  # Example:
  #
  # Let's say you have an Array of open and closed issues and want to
  # format it so the open ones are red and the closed ones are green,
  # so you can quickly see how you are doing. This can be done via
  # the colour component of "module Colours". An example can be
  # found at the end of the file here.
  # ========================================================================= #
  def format
    @ticks = @ticks.map.with_index { |tick, index|
      yield(tick, @original_numbers[index], index)
    }
    self # Return Sparky itself too.
  end

  # ========================================================================= #
  # === normalize_numbers
  #
  # Returns the normalized equivalent of a given list
  #
  #   normalize_numbers([3, 4, 7])
  #   # => [0, 1, 4]
  #
  # @return [Fixnum] the normalized equivalent of the given +_numbers+
  # ========================================================================= #
  def normalize_numbers(i)
    numbers = i.map(&:to_i)
    min = numbers.min
    numbers.map { |n| n - min }
  end

  # ========================================================================= #
  # === step_height
  #
  # @return [Float] the numerical "height" of a single bar "step"
  # ========================================================================= #
  def step_height(i)
    min, max = i.minmax
    actual_height = max - min
    step = actual_height / steps.to_f
    if step == 0
      1
    else
      step
    end
  end

  # ========================================================================= #
  # === steps
  #
  # @return [Fixnum] the count of steps between the smallest and
  # highest bar.
  # ========================================================================= #
  def steps
    PROGRESSION_BLOCKS.size - 1
  end

  # ========================================================================= #
  # === to_s
  #
  # @param sep [String, nil] separator used to join the bars of the
  # sparkline.
  #
  # @return [String] the sparkline, seperated by +sep+ (defaults to '')
  # ========================================================================= #
  def to_s(sep = nil)
    @ticks.join(sep || DEFAULT_SEPARATOR)
  end

  # ========================================================================= #
  # === lightgreen
  # ========================================================================= #
  def lightgreen(i)
    Colourslightgreen(i)
  end

end

  # ========================================================================= #
  # === AsciiParadise.sparky
  #
  # This method will return a UTF-8 "sparky" line.
  #
  # Invocation example:
  #
  #   array = [3, 33]
  #   AsciiParadise.sparky(array)
  #
  # Next, let's show an example with a block passed to the method:
  #
  # Let's say you have your list of open and closed issues, but you
  # want to format it so the open ones are red and the closed ones are
  # green, so you can quickly see how you are doing. Let's further
  # suppose you use a gem that adds a #color method to String
  # for ANSI coloring.
  #
  #   line = AsciiParadise.sparky(list) { |tick, count, index|
  #     if index == 0
  #       tick.color(:red)
  #     else
  #       tick.color(:green)
  #     end
  #   }
  #   line
  #
  # @param numbers [Array<String, Fixnum, Float>] the numbers to be rendered
  # @param &block [Proc] optional, can be used to format the rendered string
  #
  # @return [String]
  # ========================================================================= #
  def self.sparky(numbers, &block)
    new_sparky_instance = Sparky.new(numbers)
    # ======================================================================= #
    # Handle blocks given to this method.
    # ======================================================================= #
    if block
      new_sparky_instance.format(&block)
    end
    new_sparky_instance.to_s
  end; self.instance_eval { alias sparkline sparky } # === AsciiParadise.sparkline

  # ========================================================================= #
  # === return_checkmark_symbol
  # ========================================================================= #
  def return_checkmark_symbol
    CHECKMARK_SYMBOL
  end

end

if __FILE__ == $PROGRAM_NAME
  require 'colours/colours_e_autoinclude.rb'
  array = [8, 33]
  sparkline = AsciiParadise::Sparky.new(array)
  sparkline.format { |tick, count, index|
    if index == 0
      sparkline.lightgreen(tick)
    else
      Colours.darkgreen(tick)
    end
  }
  puts sparkline.to_s
end