#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === AsciiParadise::Box
#
# Usage example:
#
#   AsciiParadise::Box.new
#
# =========================================================================== #
# require 'ascii_paradise/static_ascii/box.rb'
# =========================================================================== #
require 'ascii_paradise/base/base.rb'

module AsciiParadise

class Box < Base # === AsciiParadise::Box

  # ========================================================================= #
  # === TOKEN
  # ========================================================================= #
  TOKEN = '=' # Which token to use.

  # ========================================================================= #
  # === DEFAULT_TITLE
  #
  # The default title to use will be denoted here. This title will
  # appear on top of the ASCII-box.
  # ========================================================================= #
  DEFAULT_TITLE = 'Default Title'

  # ========================================================================= #
  # === MAXIMUM_ALLOWED_LENGTH
  # ========================================================================= #
  MAXIMUM_ALLOWED_LENGTH = 80

  # ========================================================================= #
  # === DEFAULT_INPUT
  # ========================================================================= #
  DEFAULT_INPUT = {
    'foo'   => 'bar',
    'cat'   => 'tom',
    'mouse' => 'jerry',
    'donaudampffschif' => 'piggie'
  }

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = nil,
      run_already = true
    )
    reset
    set_input(i)
    if block_given?
      yielded = yield
      if yielded == :no_title
        no_title
      elsif yielded.is_a? Hash # Handle hash input.
        if yielded.has_key? :title
          set_title(yielded.delete(:title))
        end
      end
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset # (reset tag)
    super()
    reset_result
    @line_width = 80 # How long each line should be at maximum.
    @show_title = true # if true then we show the title in the header.
    set_title
  end

  # ========================================================================= #
  # === reset_result
  # ========================================================================= #
  def reset_result
    @result = ''.dup
  end

  # ========================================================================= #
  # === pad
  #
  # This method will pad input properly with a leading and a trailing '#'.
  # ========================================================================= #
  def pad(i)
    "# #{i} #"
  end

  # ========================================================================= #
  # === result?
  # ========================================================================= #
  def result?
    @result
  end; alias result result? # === result

  # ========================================================================= #
  # === no_title
  # ========================================================================= #
  def no_title
    @show_title = false
  end

  # ========================================================================= #
  # === set_input
  #
  # The input should be a Hash. It may however had also be an Array,
  # in which case we will convert it.
  # ========================================================================= #
  def set_input(i = DEFAULT_INPUT)
    i = DEFAULT_INPUT if i.nil?
    if i.is_a? Array # Check for commandline-arguments here.
      joined = i.join(' ')
      if joined.include? '--title'
        joined =~ /--title (\w+)/
        set_title($1.to_s.dup)
        i = nil # reset here
      end 
    end
    if i.is_a? String
      i = { i => '' }
    elsif i.is_a? Array # Convert Arrays into a Hash.
      i = Hash[*i]
    end
    i = sanitize_hash(i) if i
    @input = i
    determine_longest_entry
  end

  # ========================================================================= #
  # === sanitize_hash
  #
  # Key and value must be a string value.
  # ========================================================================= #
  def sanitize_hash(hash)
    _ = {}
    hash.each_pair {|key, value|
      _[key.to_s] = value.to_s.chomp
    }
    _
  end

  # ========================================================================= #
  # === input?
  # ========================================================================= #
  def input?
    @input
  end

  # ========================================================================= #
  # === token?
  # ========================================================================= #
  def token?
    TOKEN
  end

  # ========================================================================= #
  # === width?
  # ========================================================================= #
  def width?
    @line_width
  end

  # ========================================================================= #
  # === center_this_string
  # ========================================================================= #
  def center_this_string(i, use_this_token = ' ')
    i.center(width?, use_this_token)
  end; alias center center_this_string # === center

  # ========================================================================= #
  # === add_body
  # ========================================================================= #
  def add_body
    input?.each_pair {|key, value|
      string = key.ljust(@max)+' : '+value.ljust(@max)
      add pad(center(string))
    } if input?
  end

  # ========================================================================= #
  # === add
  # ========================================================================= #
  def add(i)
    @result << i+N
  end

  # ========================================================================= #
  # === draw_footer
  # ========================================================================= #
  def draw_footer
    add pad(token? * (width?))
  end

  # ========================================================================= #
  # === draw_header                                              (header tag)
  #
  # This method willd raw the title.
  #
  # If Konsole is available then we will also colourize the input.
  # ========================================================================= #
  def draw_header(optional_title = title?)
    optional_title = ''.dup if optional_title.nil?
    optional_title = " #{optional_title} " # Pad it here.
    optional_title = ''.dup unless @show_title  # Reset here when we won't show the title.
    optional_title = optional_title.center(20, token?)
    add pad( center(optional_title, token?) )
  end

  # ========================================================================= #
  # === determine_longest_entry
  #
  # This method will determine the longest entry.
  # ========================================================================= #
  def determine_longest_entry
    if input?
      @max = 0
      input?.each_pair {|key, value|
        if key.size > @max
          @max = key.size
        elsif value.size > @max
          @max = value.size
        end
      }
      if @max > MAXIMUM_ALLOWED_LENGTH
        @max = MAXIMUM_ALLOWED_LENGTH
      end
    end
  end

  # ========================================================================= #
  # === title?
  # ========================================================================= #
  def title?
    @title
  end

  # ========================================================================= #
  # === show_result
  # ========================================================================= #
  def show_result
    e result?
  end; alias report show_result # === report

  # ========================================================================= #
  # === set_title
  #
  # Set the title to use for the ASCII Box. The title will appear on top
  # of the box.
  # ========================================================================= #
  def set_title(i = DEFAULT_TITLE)
    @title = i
  end

  # ========================================================================= #
  # === colourize_title
  # ========================================================================= #
  def colourize_title(use_this_colour = :slateblue)
    @title = ::Colours.send(use_this_colour, @title)
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    reset_result
    draw_header title?
    add_body # Add the body of the box.
    draw_footer
  end

  # ========================================================================= #
  # === AsciiParadise::Box[]
  # ========================================================================= #
  def self.[](i)
    new(i).result
  end

end

  # ========================================================================= #
  # === AsciiParadise.ascii_box 'Hello World!'
  # ========================================================================= #
  def self.ascii_box(msg_content = 'Hello World!')
    AsciiParadise::Box.new(msg_content).report
  end; self.instance_eval { alias box ascii_box } # === AsciiParadise.box

  # ========================================================================= #
  # === AsciiParadise.Box
  # ========================================================================= #
  def self.Box
    AsciiParadise::Box
  end

end

if __FILE__ == $PROGRAM_NAME
  require 'cliner'
  require 'colours/colours_e_autoinclude.rb'
  array = %w(
    one two three four five six seven eight nine ten
  )
  if ARGV.empty?
    # Default for when there are no arguments given to this class.
    e; e Colours.slateblue('Testing AsciiParadise::Box next.'); e
    AsciiParadise::Box.new(
      { 'foo' => 'bar', 'cat'   => 'tom', 'mouse' => 'jerry', 'dog' => 'spike' }
    ).show_result; e
    hash = Hash[*array]
    AsciiParadise::Box.new(hash) { :no_title }.show_result; e # No title here.
    AsciiParadise::Box.new(AsciiParadise::Box::DEFAULT_INPUT) { :no_title }.show_result; e
    AsciiParadise::Box.new(AsciiParadise::Box::DEFAULT_INPUT) {{ title: 'yo there this is my title' }}.show_result; e
    _ = AsciiParadise::Box.new(AsciiParadise::Box::DEFAULT_INPUT) {{ title: 'yo there this is my title' }}
    _.colourize_title :lightgreen
    _.run
    _.show_result
    e
    AsciiParadise::Box.new(AsciiParadise::Box::DEFAULT_INPUT) {{
      title: ::Colours.steelblue('Hello world title')
    }}.show_result
    e
    e
  else
    AsciiParadise::Box.new(ARGV).show_result; e
  end
end # ascii_box