#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === AsciiParadise::Circle
#
# Solution to Ruby Quiz #166 - Drawing Circle
# =========================================================================== #
# Usage examples:
#
#   AsciiParadise::Circle.new(5).to_s
#   AsciiParadise::Circle.new(5, 2).to_s
#   AsciiParadise::Circle.new(5, 2, 'x').to_s
#
# Objects of class Circle draw circles to stdout.
#
# The aspect ratio correction is actually made drawing an ellipse with
# semimajor axis (a) equals to the given circle radius and semiminor
# axis (b) equals to a / aspect_ratio.
#
# The Circle class is responsible to:
# 
#  * initialize a Circle object with the given radius, aspect ratio
#    and drawing char
#  * initializing a canvas
#  * drawing the circle on its internal canvas 
#  * converting the canvas to string for output on stdout
# =========================================================================== #
require 'pp'

module AsciiParadise

class Circle # AsciiParadise::Circle

  # ========================================================================= #
  # cx, cy are the coordinates of the circle's center.
  # ========================================================================= #
  attr_reader :cx
  attr_reader :cy
  attr_reader :radius  
  # ========================================================================= #
  # w, h are width and height of the canvas
  # ========================================================================= #
  attr_reader :w, :h  

  # ========================================================================= #
  # === initialize
  #
  # Initialize a Circle object passing a value for radius, aspect
  # ratio and drawing character.
  # ========================================================================= #
  def initialize(
      radius,
      aspect_ratio = 1.0,
      char         = '#'
    )    
    @radius = radius.to_i
    @aspect_ratio = aspect_ratio.to_f
    @char   = char
    @radius = 10 if @radius <= 0
    fail 'Error: aspect ratio must be > 0' if @aspect_ratio <= 0        
    # ======================================================================= #
    # a is the semimajor axis of the ellipse and is equal to the given
    # radius
    # ======================================================================= #
    @a = @radius
    # ======================================================================= #
    # b is the semiminor axis of the ellipse and is calculated from a
    # and the given aspect ratio
    # ======================================================================= #
    @b = (@a / @aspect_ratio).round    
    # ======================================================================= #
    # calculate the size of the canvas
    # ======================================================================= #
    @w, @h = @a * 2 + 1, @b * 2 + 1    
    # ======================================================================= #
    # The center coordinates correspond to the size of semiaxis.
    # ======================================================================= #
    @cx, @cy = @a, @b
    # ======================================================================= #    
    # initialize the canvas with spaces
    # ======================================================================= #
    @canvas = Array.new(@w * @h, ' ')    
    # ======================================================================= #
    # draw ellipse on the canvas
    # ======================================================================= #
    draw_ellipse(@a, @b)
  end

  # ========================================================================= #
  # === point
  #
  # Draw the given character on canvas to the given coordinates.
  # ========================================================================= #
  def point(x, y)
    @canvas[y * @w + x] = @char
  end
  
  # ========================================================================= #
  # === plot_four_points
  #
  # Translates and mirrors point (x, y) in the quadrants taking advantage
  # of the simmetries in the ellipse. Thus, for a given point (x, y) the
  # method plot three other points in the remaining quadrants.
  # ========================================================================= #
  def plot_four_points(x, y)
    point(@cx + x, @cy + y)
    point(@cx - x, @cy + y)
    point(@cx + x, @cy - y)
    point(@cx - x, @cy - y)
  end
  
  # ========================================================================= #
  # === draw_ellipse
  #
  # Draw an ellipse on canvas. This method implements a Bresenham
  # based algorithm by John Kennedy:
  #
  #   http://homepage.smc.edu/kennedy_john/BELIPSE.PDF
  #
  # The method calculates two set of points in the first quadrant.
  #
  # The first set starts on the positive x axis and wraps in a
  # counterclockwise direction until the tangent line slope is equal
  # to -1.  The second set starts on the positive y axis and wraps in
  # a clockwise direction until the tangent line slope is equal to -1.
  # ========================================================================= #
  def draw_ellipse(a, b)
    a_square = 2 * a ** 2
    b_square = 2 * b ** 2
    draw_first_set( a, b, a_square, b_square)
    draw_second_set(a, b, a_square, b_square)
  end

  # ========================================================================= #
  # === draw_first_set
  #
  # The method increments y and decides when to decrement x testing
  # the sign of a function.  In this case, the decision function is
  # (2*ellipse_error+x_change) and its value is calculated
  # iteratively.
  # ========================================================================= #
  def draw_first_set(a, b, a_square, b_square)
    x, y = a, 0
    x_change, y_change = b_square / 2 * (1 - 2 * a), a_square / 2
    stopping_x, stopping_y = b_square * a, 0
    ellipse_error = 0

    while (stopping_x >= stopping_y) do
      plot_four_points(x, y)
      y += 1
      stopping_y += a_square
      ellipse_error += y_change
      y_change += a_square
      if (2 * ellipse_error + x_change) > 0
        x -= 1
        stopping_x -= b_square
        ellipse_error += x_change
        x_change += b_square
      end
    end
  end

  # ========================================================================= #
  # === draw_second_set
  #
  # The method increments x and decides when to decrement y testing
  # the sign of a function.  In this case, the decision function is
  # (2*ellipse_error+y_change) and its value is calculated
  # iteratively.
  # ========================================================================= #
  def draw_second_set(a, b, a_square, b_square)
    x, y = 0, b
    x_change, y_change = b_square / 2, a_square / 2 * (1 - 2 * b)
    stopping_x, stopping_y = 0, a_square * b
    ellipse_error = 0
    
    while (stopping_x <= stopping_y) do
      plot_four_points(x, y)
      x += 1
      stopping_x += b_square
      ellipse_error += x_change
      x_change += b_square
      if (2 * ellipse_error + y_change) > 0
        y -= 1
        stopping_y -= a_square
        ellipse_error += y_change
        y_change += a_square
      end
    end
  end

  # ========================================================================= #
  # === to_s
  #
  # Return the circle-data on as a String.
  # ========================================================================= #
  def to_s
    result = ''.dup
    (0..@h - 1).each { |line|
      result << @canvas[line * @w..line * @w + @w - 1].join << "\n"
    }
    result
  end

  # ========================================================================= #
  # === output
  # ========================================================================= #
  def output
    print to_s
  end; alias display output # === display

  # ========================================================================= #
  # === canvas?
  #
  # @canvas is a linear Array that is initially filled with spaces.
  # ========================================================================= #
  def canvas?
    @canvas
  end; alias canvas canvas? # === canvas

end

  # ========================================================================= #
  # === AsciiParadise.ascii_circle
  #
  # We draw an ascii circle here. We can pass the radius as its first
  # argument.
  # ========================================================================= #
  def self.ascii_circle(radius = 5)
    AsciiParadise::Circle.new(radius, 1.0, '#').display
  end

  # ========================================================================= #
  # === AsciiParadise.Circle
  # ========================================================================= #
  def self.Circle
    AsciiParadise::Circle
  end

end

# =========================================================================== #
# Usage examples:
#   ascii_circle.rb 7 #=> print out a circle of radius 7
#   ascii_circle.rb 7 1.8 #=> print out a circle of radius 7 and aspect ratio 1.8
#   ascii_circle.rb 7 1.8 x #=> print out a circle of radius 7 and aspect ratio 1.8
#   ascii_circle.rb 15
# =========================================================================== #
if __FILE__ == $PROGRAM_NAME
  AsciiParadise::Circle.new(ARGV.first, ARGV[1] || 1.0, ARGV[2] || '#').output
end # ascii_circle