#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'ascii_paradise/toplevel_methods/available_ascii_components.rb'
# AsciiParadise.available_ascii_components?
# =========================================================================== #
require 'ascii_paradise/project/project.rb'
require 'ascii_paradise/requires/require_animations_files.rb'
require 'ascii_paradise/toplevel_methods/available_ascii_components_options.rb'
require 'ascii_paradise/toplevel_methods/show_available_ascii_components.rb'

module AsciiParadise

  # ========================================================================= #
  # === AsciiParadise.available_ascii_components?
  #
  # This will return a sorted Array which will have all the ascii
  # components that are available.
  #
  # This Array will look like so:
  #
  #   ["AirplaneOverForest", "AnimatedCigarette"] # And so on.
  # 
  # ========================================================================= #
  def self.available_ascii_components?
    target = ::AsciiParadise.animations_directory?
    array = []
    Dir["#{target}*.rb"].each {|file|
      _ = File.basename(file).
          gsub(/ascii_/,'').
          sub(/\.rb$/,'').capitalize
      # ===================================================================== #
      # Camelcase it:
      # ===================================================================== #
      if _.include? '_'
        _ = _.split('_').map {|entry| entry.capitalize }.join
      end
      array << _
    }
    array.sort # Must sort it alphabetically.
  end; self.instance_eval { alias available_components  available_ascii_components? } # === AsciiParadise.available_components
       self.instance_eval { alias available_components? available_ascii_components? } # === AsciiParadise.available_components?

  # ========================================================================= #
  # === AsciiParadise.n_animated_ascii_components
  #
  # This method will return a number, which tells us how many ASCII
  # components are registered in this project.
  # ========================================================================= #
  def self.n_animated_ascii_components
    ::AsciiParadise.available_ascii_components?.size
  end; self.instance_eval { alias n_animated_ascii_components? n_animated_ascii_components } # === AsciiParadise.n_animated_ascii_components?

  # ========================================================================= #
  # === AsciiParadise.return_all_animated_ascii_components
  #
  # This method will return all AsciiParadise components that can be
  # animated.
  #
  # For now we assume that this will only work for all those components
  # that are a subclass of class AsciiParadise::Animation.
  #
  # Usage example:
  #
  #   x = AsciiParadise.return_all_animated_ascii_components
  #
  # ========================================================================= #
  def self.return_all_animated_ascii_components
    # ======================================================================= #
    # First obtain the available candidates:
    # ======================================================================= #
    possible_candidates = available_ascii_components?
    # ======================================================================= #
    # Turn these into real constants, via const_get(). In order for this
    # to work, the whole dataset has to be initialized for these individual
    # classes.
    # ======================================================================= #
    possible_candidates.map! {|entry|
      if ::AsciiParadise.const_defined? entry.to_sym
        entry = ::AsciiParadise.const_get(entry)
      end
      entry
    }
    possible_candidates.compact!
    # ======================================================================= #
    # Next select the proper entries that contain the dataset.:
    # ======================================================================= #
    possible_candidates.select! {|entry|
      entry.is_animated?
    }
    # ======================================================================= #
    # Now we have the entries that we want to have, thus we can return them.
    # ======================================================================= #
    return possible_candidates
  end

end

if __FILE__ == $PROGRAM_NAME
  require 'colours/autoinclude'
  e 'The available components are:'
  e
  pp AsciiParadise.available_ascii_components?
  e
  e 'The AsciiParadise project has '+
     sfancy(AsciiParadise.n_animated_ascii_components)+' components.'
  e
  AsciiParadise.show_available_ascii_components
  e
  p AsciiParadise.return_all_animated_ascii_components
  e
end