#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === AsciiParadise::RotatingDNA
# =========================================================================== #
require 'ascii_paradise/animation/animation.rb'

module AsciiParadise

class RotatingDNA < Animation

  # ========================================================================= #
  # === THRESHOLD_N_RUNS_ALLOWED
  # ========================================================================= #
  THRESHOLD_N_RUNS_ALLOWED = 250 # How many runs we will do.

  # ========================================================================= #
  # === UPDATE_SPEED
  #
  # How fast the DNA will rotate, at each interval.
  # ========================================================================= #
  UPDATE_SPEED = 0.05

  # ========================================================================= #
  # === DNA_X_Y_MOVEMENT_SPEED
  #
  # The higher this number is, the faster the DNA will "rotate".
  # 2 is my default.
  # ========================================================================= #
  DNA_X_Y_MOVEMENT_SPEED = 2

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      use_this_runmode = :dna,
      run_already      = true
    )
    super()
    reset
    register_sigint
    set_runmode(use_this_runmode)
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    set_runmode # Default mode is :dna. The other mode is :star
    @n_runs = 0
  end

  # ========================================================================= #
  # === draw_star_frame
  # ========================================================================= #
  def draw_star_frame(
      clear,
      time_scale,
      horiz_scale,
      vert_scale,
      curve_length,
      pair_space,
      pair_offset = pair_space / 2,
      char = '*'
    )
    frame = Array.new(vert_scale){Array.new(curve_length, ' ')}
    t = Time.now.to_f
    sin_start = time_scale*t
    half_vs = vert_scale / 2
    (0...curve_length).each{|i| 
      if (i+pair_offset) % pair_space < 1
        v = half_vs * Math.sin(sin_start+horiz_scale*i)
        p1 = (half_vs + v)
        p2 = (half_vs - v)
        frame[p1][i] = char
        frame[p2][i] = char
      end
    }
    print_clear
    e frame.map {|line| line.join }
  end

  # ========================================================================= #
  # === draw_dna_frame
  #
  # Draw the DNA frame. This is the main powerhorse of this class.
  # ========================================================================= #
  def draw_dna_frame(
      time_scale,   horiz_scale, vert_scale, 
      curve_length, pair_space, 
      pair_offset = pair_space/2, char = '*'
    )
    frame = Array.new(vert_scale) { Array.new(curve_length, ' ') }
    time_now  = Time.now.to_f # Get the start-time.
    sin_start = time_scale * time_now
    half_vs   = vert_scale/2
    (0...curve_length).each { |i|
      v = half_vs * Math.sin(sin_start+horiz_scale*i)
      p1 = half_vs + v
      p2 = half_vs - v
      frame[p1][i] = char
      frame[p2][i] = char
      if (i + pair_offset) % pair_space < 1
        p1,p2 = [p1, p2].sort
        (p1.ceil...p2.ceil).each { |y| frame[y][i] = char }
      end
    }
    clear
    e frame.map {|line| line.join }
  end

  # ========================================================================= #
  # === increment_counter
  # ========================================================================= #
  def increment_counter
    @n_runs += 1
  end

  # ========================================================================= #
  # === print_clear
  # ========================================================================= #
  def print_clear
    print `clear`
  end; alias clear print_clear # === clear

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    loop {
      case runmode?
      when :dna, :default
        draw_dna_frame(DNA_X_Y_MOVEMENT_SPEED, 0.1, 11, 60, 8, 4)
      else
        draw_star_frame(`clear`,2,0.1,11,60,4,2)
      end
      sleep UPDATE_SPEED
      increment_counter
      exit if @n_runs > THRESHOLD_N_RUNS_ALLOWED
    }
  end

  # ========================================================================= #
  # === set_runmode
  # ========================================================================= #
  def set_runmode(use_this_runmode = :dna)
    @runmode = use_this_runmode
  end

  # ========================================================================= #
  # === runmode?
  #
  # @runmode right now can be :dna, or :star. :star will be the default.
  # ========================================================================= #
  def runmode?
    @runmode
  end; alias draw_what? runmode? # === draw_what?

  # ========================================================================= #
  # === AsciiParadise::RotatingDNA[]
  #
  # The runmodes can be:
  #
  #   AsciiParadise::RotatingDNA[:dna]
  #   AsciiParadise::RotatingDNA[:star]
  #
  # ========================================================================= #
  def self.[](runmode = :dna)
    new(runmode)
  end

end

# =========================================================================== #
# === AsciiParadise.rotating_dna
# =========================================================================== # 
def self.rotating_dna(runmode = :dna)
  AsciiParadise::RotatingDNA[runmode]
end

# =========================================================================== #
# === AsciiParadise.RotatingDna
# =========================================================================== # 
def self.RotatingDna
  AsciiParadise::RotatingDNA
end; self.instance_eval { alias RotatingDNA RotatingDna } # === AsciiParadise.RotatingDNA

end

if __FILE__ == $PROGRAM_NAME
  AsciiParadise.rotating_dna
end # rotating_dna